import functools
import threading
from typing import *
from abc import ABC, abstractmethod


def get_not_none(obj: Any, default: Union[Any, Callable[[], Any]], call_callables=True):
    return obj if obj is not None else default if not call_callables or not callable(default) else default()


def synchronized(wrapped):
    lock = threading.Lock()

    @functools.wraps(wrapped)
    def _wrap(*args, **kwargs):
        with lock:
            return wrapped(*args, **kwargs)

    return _wrap



import time

from ... import Event, TriggererEventSource
from ..._private._utilities import *


class PeriodicEventSource(TriggererEventSource):
    def __init__(self,
                 generator_function: Callable[[], Event],
                 sec_to_regenerate: int,
                 # TODO add start and end time in the format of cron schedule expression
                 **kwargs):
        super().__init__(**kwargs)
        self.generator_function = generator_function
        self.sec_to_regenerate = sec_to_regenerate
        self.consuming: bool = False

    def start_generating(self) -> None:
        self.consuming = True
        threading.Thread(target=self._repetitively_generate).start()

    def _repetitively_generate(self):
        last_run: float = 0
        while self.consuming:
            try:
                time.sleep(1)
                now = time.time()
                if (last_run // self.sec_to_regenerate) != (now // self.sec_to_regenerate):
                    event: Event = self.generator_function()
                    last_run = now
                    self.process_generated_event(event)
            except Exception as e:
                self.logger.exception(f'Failed to generate event at time: "{str(e)}"')

    def stop_generating(self) -> None:
        self.consuming = False



from collections import deque
from multiprocessing.pool import ThreadPool
from time import sleep

from ..config import PipelineConfig
from ._bus import *
from ._elements import *
from ._resources import get_banner
from ..events import ProcedureEvent, PipelineStartEvent, PipelineStopEvent


class Pipeline:
    def __init__(self,
                 sources: List[EventSource] = None,
                 sinks: Dict[EventSink, List[Type]] = None,
                 transformers: Dict[EventTransformer, List[Type]] = None,
                 other_generators: List[EventGenerator] = None,
                 other_listeners: Dict[EventListener, List[Type]] = None,
                 config: PipelineConfig = None
                 ):
        sources = get_not_none(sources, [])
        sinks = get_not_none(sinks, {})
        transformers = get_not_none(transformers, {})
        other_generators = get_not_none(other_generators, [])
        other_listeners = get_not_none(other_listeners, {})

        self._procedure_informer = ProcedureInformer()

        self.event_listeners: List[Tuple[EventListener, List[Type]]] = \
            list(other_listeners.items()) + list(sinks.items()) + list(transformers.items())
        self.event_generators: List[EventGenerator] = \
            [self._procedure_informer] + other_generators + sources + list(transformers.keys())

        self.bus: EventBus = EventBus()
        for generator in self.event_generators:
            generator.event_handler = InputLoggerEventBus(generator, self.bus)
        for listener in self.event_listeners:
            for t in listener[1]:
                self.bus.add_listener(listener[0], t)

        self.config: PipelineConfig = get_not_none(config, lambda: PipelineConfig())
        self.logger: Logger = self.config.logger

        all_pipeline_elements: Set[InternalPipelineElement] = set(self.event_generators).union(
            [x[0] for x in self.event_listeners])
        for element in all_pipeline_elements:
            element.set_data_store(self.config.data_store)
            if element.logger is None:
                element.set_logger(self.logger.getChild(element.name))

        # self.use_ray = (ray_config is not None)
        # self.ray_config = ray_config
        # if self.use_ray:
        #     # ray.init(address='ray://?')  # server

    def start(self):
        print(get_banner())
        # self.logger.info("Starting Pipeline")
        self._procedure_informer.inform_event(PipelineStartEvent())
        # if self.use_ray:
        #     ray.get(start_pipeline_module_with_ray.remote(self))
        # else:
        #     self.start_directly()
        try:
            self._start_event_sources()
        except KeyboardInterrupt:
            self.stop()

    def stop(self):
        # self.logger.info("Stopping Pipeline")
        self._procedure_informer.inform_event(PipelineStopEvent())
        self._stop_event_sources()

    def _start_event_sources(self):
        self._start_triggerer_sources()
        self._start_triggering_triggerable_sources()

    def _stop_event_sources(self):
        self._stop_triggerer_sources()
        self._stop_triggering_triggerable_sources()

    def _start_triggerer_sources(self):
        for generator in self.event_generators:
            if isinstance(generator, TriggererEventSource):
                generator.start_generating()

    def _stop_triggerer_sources(self):
        for generator in self.event_generators:
            if isinstance(generator, TriggererEventSource):
                generator.stop_generating()

    def _start_triggering_triggerable_sources(self):
        triggerable_sources = \
            [generator for generator in self.event_generators if isinstance(generator, TriggerableEventSource)]
        priorities = list(set(map(lambda x: x.priority, triggerable_sources)))
        priorities.sort()
        prioritised_sources = [deque([source for source in triggerable_sources if source.priority == priority])
                               for priority in priorities]
        self.triggering_triggerable_sources = True
        max_queue_size = 2 * self.config.source_triggerers_thread_count
        with ThreadPool(processes=self.config.source_triggerers_thread_count) as p:
            while self.triggering_triggerable_sources:
                # noinspection PyProtectedMember
                while p._inqueue.qsize() > max_queue_size:
                    sleep(0.001)
                event_generator_function: Optional[Callable[[], None]] = None
                for priority_sources in prioritised_sources:
                    for i in range(len(priority_sources)):
                        event_generator_function = priority_sources[i].get_event_generator_function()
                        if event_generator_function is not None:
                            priority_sources.rotate(-i - 1)
                            break
                    if event_generator_function is not None:
                        break
                if event_generator_function is not None:
                    p.apply_async(event_generator_function)

    def _stop_triggering_triggerable_sources(self):
        self.triggering_triggerable_sources = False


# @ray.remote
# def start_pipeline_module_with_ray(pipeline_module: Pipeline):
#     pipeline_module.start_directly()


class ProcedureInformer(EventGenerator):
    def __init__(self, **kwargs):
        super().__init__(output_flow_log_level=logging.INFO, **kwargs)

    def inform_event(self, event: ProcedureEvent):
        self.event_handler.handle_event(event)



from ... import Server


class KafkaTopic:
    def __init__(self,
                 server: Server,
                 topic_id: str
                 ):
        self.server = server
        self.topic_id = topic_id


class ProducerConfig:
    def __init__(self, **kwargs):
        """
        :kwargs args for https://kafka-python.readthedocs.io/en/master/apidoc/KafkaProducer.html
        """
        self.producer_kwargs = kwargs



from ._utilities import *
from ._concepts import EventHandler, Event
from ._elements import EventListener, InternalPipelineElement


class EventBus(EventHandler):
    def __init__(self):
        self.listeners: Dict[Type, List[EventListener]] = dict()

    def add_listener(self, listener: EventListener, type_to_listen: Type) -> None:
        if type_to_listen not in self.listeners:
            self.listeners[type_to_listen] = [listener]
        else:
            self.listeners[type_to_listen].append(listener)

    def handle_event(self, event_to_handle: Event):
        event_type: Type = type(event_to_handle)
        for listed_type in self.listeners:
            if issubclass(event_type, listed_type):
                for event_listener in self.listeners[listed_type]:
                    # noinspection PyProtectedMember
                    event_listener.logger.log(event_listener.input_flow_log_level,
                                              f'<<< {event_to_handle.__class__.__name__} ({event_to_handle._repr()})')
                    try:
                        event_listener.process_event(event_to_handle)
                    except Exception as e:
                        event_listener.logger.exception(f'Failed to process event: {str(e)}')


class InputLoggerEventBus(EventHandler):
    def __init__(self, raiser: InternalPipelineElement, bus: EventBus):
        self.raiser = raiser
        self.bus = bus

    def handle_event(self, event_to_handle: Event):
        # noinspection PyProtectedMember
        self.raiser.logger.log(self.raiser.output_flow_log_level,
                               f'>>> {event_to_handle.__class__.__name__} ({event_to_handle._repr()})')
        self.bus.handle_event(event_to_handle)



from ... import Server


class SocketConfig:
    def __init__(self,
                 server: Server,
                 max_reconnect_timeout: float = 5,
                 initial_reconnect_timeout: float = 0.1,
                 **kwargs):
        """
        :kwargs args for https://websockets.readthedocs.io/en/7.0/api.html#websockets.client.connect
        """
        self.server = server
        self.max_reconnect_timeout = max_reconnect_timeout
        self.initial_reconnect_timeout = initial_reconnect_timeout
        self.connection_kwargs = kwargs
